"""Abstract base recommender class."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Set as AbstractSet

    import polars as pl

GameKeyType = TypeVar("GameKeyType")
UserKeyType = TypeVar("UserKeyType")


class BaseGamesRecommender(ABC, Generic[GameKeyType, UserKeyType]):
    """Abstract base games recommender class."""

    @property
    @abstractmethod
    def known_games(self) -> AbstractSet[GameKeyType]:
        """IDs of all known games."""

    @property
    def num_games(self) -> int:
        """Number of known games."""
        return len(self.known_games)

    @property
    @abstractmethod
    def rated_games(self) -> AbstractSet[GameKeyType]:
        """IDs of all rated games."""

    @property
    @abstractmethod
    def known_users(self) -> AbstractSet[UserKeyType]:
        """IDs of all known users."""

    @property
    def num_users(self) -> int:
        """Number of known users."""
        return len(self.known_users)

    @abstractmethod
    def recommend(
        self,
        users: Iterable[UserKeyType],
        **kwargs,
    ) -> pl.DataFrame:
        """Recommend games for given users."""

    @abstractmethod
    def recommend_as_numpy(
        self,
        users: Iterable[UserKeyType],
        games: Iterable[GameKeyType],
    ) -> np.ndarray:
        """Recommend games for given users and games as a numpy array."""

    @abstractmethod
    def recommend_group(
        self,
        users: Iterable[UserKeyType],
        **kwargs,
    ) -> pl.DataFrame:
        """Recommend games for given group of users."""

    @abstractmethod
    def recommend_group_as_numpy(
        self,
        users: Iterable[UserKeyType],
        games: Iterable[GameKeyType],
    ) -> np.ndarray:
        """Recommend games for given group of users and games as a numpy array."""

    @abstractmethod
    def recommend_similar(
        self,
        games: Iterable[GameKeyType],
        **kwargs,
    ) -> pl.DataFrame:
        """Recommend games similar to the given ones."""

    @abstractmethod
    def similar_games(
        self,
        games: Iterable[GameKeyType],
        **kwargs,
    ) -> pl.DataFrame:
        """Find games similar to the given ones."""

    def recommend_random_games_as_numpy(
        self,
        users: Iterable[UserKeyType],
        games: Iterable[GameKeyType],
        *,
        num_games: int = 1,
        random_seed: int | None = None,
    ) -> np.ndarray:
        """
        Select random games based on recommendations for a given group of
        users and games as a numpy array.
        """

        users = list(users)
        games = list(games)

        if not games:
            return np.array([])

        # TODO: allow for different weighting schemes
        weights = self.recommend_group_as_numpy(users, games).reshape(-1)
        weights = np.exp(weights)

        rng = np.random.default_rng(seed=random_seed)
        return rng.choice(
            a=games,  # type: ignore[arg-type]
            size=min(num_games, len(games)),
            replace=False,
            p=weights / weights.sum(),
        )



"""Light recommender model, without the heavy Turi Create dependency."""

from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np

from board_game_recommender.abc import BaseGamesRecommender
from board_game_recommender.baseline import dataframe_from_scores

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Set as AbstractSet
    from typing import Any, Self

    import polars as pl


@dataclass(frozen=True)
class CollaborativeFilteringData:
    """Labels, vectors and matrices for linear collaborative filtering models."""

    intercept: float
    users_labels: np.ndarray  # (num_users,)
    users_linear_terms: np.ndarray  # (num_users,)
    users_factors: np.ndarray  # (num_users, num_factors)
    items_labels: np.ndarray  # (num_items,)
    items_linear_terms: np.ndarray  # (num_items,)
    items_factors: np.ndarray  # (num_factors, num_items)

    def to_npz(self, file_path: Path | str) -> None:
        """Save data into an .npz file."""

        file_path = Path(file_path).resolve()
        LOGGER.info("Saving data as .npz to <%s>", file_path)
        with file_path.open(mode="wb") as file:
            np.savez(file=file, **asdict(self))
        LOGGER.info("Done saving <%s>", file_path)

    @classmethod
    def from_npz(cls, file_path: Path | str) -> Self:
        """Load data from an .npz file."""

        file_path = Path(file_path).resolve()
        LOGGER.info("Loading data as .npz from <%s>", file_path)
        with file_path.open(mode="rb") as file:
            files = np.load(file=file)
            files_dict = {
                key: float(files[key]) if key == "intercept" else files[key]
                for key in files.files
            }
            assert all(
                isinstance(key, str) and isinstance(value, (np.ndarray, float))
                for key, value in files_dict.items()
            ), "All keys must be strings and all values must be numpy arrays or floats"
            return cls(**files_dict)  # type: ignore[arg-type]


class LightGamesRecommender(BaseGamesRecommender):
    """Light recommender without Turi Create dependency."""

    _known_games: frozenset[int] | None = None
    _known_users: frozenset[str] | None = None

    def __init__(self, data: CollaborativeFilteringData) -> None:
        self.data = data

        assert data.users_factors.shape[-1] == data.items_factors.shape[0]
        num_factors = data.items_factors.shape[0]
        # TODO check other dimensions as well (num_users and num_items)

        self.intercept: float = data.intercept

        num_users = len(data.users_labels)
        self.users_labels: list[str] = list(data.users_labels)
        self.users_indexes = defaultdict(
            lambda: -1,
            zip(data.users_labels, range(num_users)),
        )
        self.users_linear_terms = np.concatenate((data.users_linear_terms, np.zeros(1)))
        self.users_factors = np.concatenate(
            (data.users_factors, np.zeros((1, num_factors))),
            axis=0,
        )

        num_items = len(data.items_labels)
        self.items_labels: list[int] = list(data.items_labels)
        self.items_indexes = defaultdict(
            lambda: -1,
            zip(data.items_labels, range(num_items)),
        )
        self.items_linear_terms = np.concatenate((data.items_linear_terms, np.zeros(1)))
        self.items_factors = np.concatenate(
            (data.items_factors, np.zeros((num_factors, 1))),
            axis=1,
        )

        LOGGER.info(
            "Loaded light recommender with %d users and %d items",
            len(self.users_labels),
            len(self.items_labels),
        )

    def to_npz(self, file_path: Path | str) -> None:
        """Save data into an .npz file."""
        self.data.to_npz(file_path)

    @classmethod
    def from_npz(
        cls,
        file_path: Path | str,
    ) -> Self:
        """Load data from an .npz file."""
        data = CollaborativeFilteringData.from_npz(file_path)
        return cls(data)

    @property
    def known_games(self) -> AbstractSet[int]:
        if self._known_games is not None:
            return self._known_games
        self._known_games = frozenset(self.items_labels)
        return self._known_games

    @property
    def rated_games(self) -> AbstractSet[int]:
        return self.known_games

    @property
    def num_games(self) -> int:
        return len(self.items_labels)

    @property
    def known_users(self) -> AbstractSet[str]:
        if self._known_users is not None:
            return self._known_users
        self._known_users = frozenset(self.users_labels)
        return self._known_users

    @property
    def num_users(self) -> int:
        return len(self.users_labels)

    def _recommendation_scores(
        self,
        *,
        users: list[str] | None = None,
        games: list[int] | None = None,
        avg_users: bool = False,
    ) -> np.ndarray:
        """Calculate recommendations scores for certain users and games."""

        if users:
            user_ids = np.array([self.users_indexes[user] for user in users])
            users_factors = self.users_factors[user_ids]
            users_linear_terms = self.users_linear_terms[user_ids].reshape(-1, 1)
        else:
            users_factors = self.users_factors[:-1, :]
            users_linear_terms = self.users_linear_terms[:-1].reshape(-1, 1)

        if avg_users:
            users_factors = users_factors.mean(axis=0).reshape(1, -1)
            users_linear_terms = users_linear_terms.mean(axis=0).reshape(1, 1)

        if games:
            game_ids = np.array([self.items_indexes[game] for game in games])
            items_factors = self.items_factors[:, game_ids]
            items_linear_terms = self.items_linear_terms[game_ids].reshape(1, -1)
        else:
            items_factors = self.items_factors[:, :-1]
            items_linear_terms = self.items_linear_terms[:-1].reshape(1, -1)

        return (
            users_factors @ items_factors  # (num_users, num_items)
            + users_linear_terms  # (num_users, 1)
            + items_linear_terms  # (1, num_items)
            + self.intercept  # (1,)
        )

    def _game_scores(
        self,
        games: list[int] | None = None,
    ) -> np.ndarray:
        """Calculate average game scores from bias terms."""

        if games:
            game_ids = np.array([self.items_indexes[game] for game in games])
            items_linear_terms = self.items_linear_terms[game_ids]
        else:
            items_linear_terms = self.items_linear_terms[:-1]

        return items_linear_terms + self.intercept

    def recommend(
        self,
        users: Iterable[str],
        **kwargs: Any,  # noqa: ARG002
    ) -> pl.DataFrame:
        """Calculate recommendations for certain users."""

        users = list(users)
        scores = self._recommendation_scores(users=users)
        return dataframe_from_scores(
            users=users,
            games=self.items_labels,
            scores=scores,
        )

    def recommend_as_numpy(
        self,
        users: Iterable[str],
        games: Iterable[int],
    ) -> np.ndarray:
        """Calculate recommendations for certain users and games as a numpy array."""

        users = list(users)
        games = list(games)

        return self._recommendation_scores(users=users, games=games)

    def recommend_group(
        self,
        users: Iterable[str],
        **kwargs: Any,  # noqa: ARG002
    ) -> pl.DataFrame:
        """Calculate recommendations for a group of users."""

        users = list(users)
        scores = (
            self._recommendation_scores(users=users, avg_users=True)
            if users
            else self._game_scores()
        )
        return dataframe_from_scores(
            users=["_all"],
            games=self.items_labels,
            scores=scores,
        )

    def recommend_group_as_numpy(
        self,
        users: Iterable[str],
        games: Iterable[int],
    ) -> np.ndarray:
        """
        Calculate recommendations for a group of users and games as a numpy array.
        """

        users = list(users)
        games = list(games)
        return (
            self._recommendation_scores(users=users, games=games, avg_users=True)
            if users
            else self._game_scores(games).reshape(1, -1)
        )

    def recommend_similar(
        self,
        games: Iterable[int],
        **kwargs,
    ) -> pl.DataFrame:
        """
        Recommend games similar to the given games based on cosine similarity
        of latent factors.
        """

        msg = "Not implemented yet"
        raise NotImplementedError(msg)

        # games = list(games)
        # game_ids = np.array([self.items_indexes[game] for game in games])
        # game_factors = self.items_factors[:, game_ids]

        # scores = cosine_similarity(game_factors, self.items_factors[:, :-1]).mean(
        #     axis=0
        # )

        # result = pd.DataFrame(index=self.items_labels, data={"score": scores})
        # result["rank"] = result["score"].rank(method="min",
        # ascending=False).astype(int)
        # result.sort_values("rank", inplace=True)

        # return result

    def similar_games(
        self,
        games: Iterable[int],
        **kwargs,
    ) -> pl.DataFrame:
        """
        Find games similar to the given games based on
        cosine similarity of latent factors.
        """

        msg = "Not implemented yet"
        raise NotImplementedError(msg)

        # games = list(games)
        # game_ids = np.array([self.items_indexes[game] for game in games])
        # game_factors = self.items_factors[:, game_ids]

        # scores = cosine_similarity(game_factors, self.items_factors[:, :-1])

        # return dataframe_from_scores(games, self.items_labels, scores)


def cosine_similarity(matrix_1: np.ndarray, matrix_2: np.ndarray) -> np.ndarray:
    """
    Calculates the cosine similarity between two matrices.

    The input matrices need to be of shape (m,n) and (m,l);
    the result shape will be (n,l).
    """

    dot_product = matrix_1.T @ matrix_2  # (n,l)
    matrix_1_norm = np.linalg.norm(matrix_1, axis=0)  # (n,)
    matrix_2_norm = np.linalg.norm(matrix_2, axis=0)  # (l,)
    outer_prod_norm = np.outer(matrix_1_norm, matrix_2_norm)  # (n,l)

    return dot_product / outer_prod_norm  # (n,l)



"""Baseline recommender models."""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, cast

import numpy as np
import polars as pl

from board_game_recommender.abc import BaseGamesRecommender

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path
    from typing import Any, Self


def dataframe_from_scores(
    *,
    users: Iterable[str],
    games: Iterable[Any],
    scores: np.ndarray,  # shape: (users, games)
) -> pl.DataFrame:
    """Creates a Pandas DataFrame out of raw recommendation scores."""

    users = tuple(users)
    games = tuple(games)

    rank_cols = (
        pl.col(f"{column}_score")
        .rank(method="min", descending=True)
        .alias(f"{column}_rank")
        for column in users
    )

    result = (
        pl.DataFrame(
            data=scores.T,
            schema=[f"{column}_score" for column in users],
        )
        .lazy()
        .with_columns(
            pl.Series("index", games),
            *rank_cols,
        )
    )

    if len(users) == 1:
        result = result.sort(f"{users[0]}_rank")

    return result.collect()


class RandomGamesRecommender(BaseGamesRecommender):
    """Random recommender."""

    def __init__(self) -> None:
        self.rng = np.random.default_rng()

    @property
    def known_games(self) -> frozenset[int]:
        return frozenset()

    @property
    def rated_games(self) -> frozenset[int]:
        return frozenset()

    @property
    def known_users(self) -> frozenset[str]:
        return frozenset()

    def _recommendation_scores(self, users: int, games: int) -> np.ndarray:
        """Random scores."""
        return self.rng.random((users, games))

    def recommend(  # type: ignore[override]
        self,
        users: Iterable[str],
        games: Iterable[int],
        **kwargs,  # noqa: ARG002
    ) -> pl.DataFrame:
        """Random recommendations for certain users."""

        users = list(users)
        games = list(games)
        scores = self._recommendation_scores(users=len(users), games=len(games))

        return dataframe_from_scores(users=users, games=games, scores=scores)

    def recommend_as_numpy(
        self,
        users: Iterable[str],
        games: Iterable[int],
    ) -> np.ndarray:
        """Random recommendations for certain users and games as a numpy array."""
        users = list(users)
        games = list(games)
        return self._recommendation_scores(users=len(users), games=len(games))

    def recommend_group(  # type: ignore[override]
        self,
        users: Iterable[str],  # noqa: ARG002
        games: Iterable[int],
        **kwargs,  # noqa: ARG002
    ) -> pl.DataFrame:
        """Random recommendations for a group of users."""

        games = list(games)
        scores = self._recommendation_scores(users=1, games=len(games))

        return dataframe_from_scores(users=["_all"], games=games, scores=scores)

    def recommend_group_as_numpy(
        self,
        users: Iterable[str],  # noqa: ARG002
        games: Iterable[int],
    ) -> np.ndarray:
        """Random recommendations for a group of users and games as a numpy array."""
        games = list(games)
        return self._recommendation_scores(users=1, games=len(games))

    def recommend_similar(self, games: Iterable[int], **kwargs) -> pl.DataFrame:
        raise NotImplementedError

    def similar_games(self, games: Iterable[int], **kwargs) -> pl.DataFrame:
        raise NotImplementedError


class PopularGamesRecommender(BaseGamesRecommender):
    """Popular games recommender."""

    id_field: str = "bgg_id"
    user_id_field: str = "bgg_user_name"
    rating_id_field: str = "bgg_user_rating"

    scores: dict[int, float]
    raw_scores: np.ndarray
    default_value: float
    game_ids: tuple[int, ...]

    _known_games: frozenset[int] | None = None

    def __init__(
        self,
        game_ids: Iterable[int],
        scores: np.ndarray,
        default_value: float | None = None,
    ) -> None:
        self.default_value = (
            default_value if default_value is not None else scores.mean()
        )
        self.raw_scores = scores
        self.game_ids = tuple(game_ids)
        self.scores = defaultdict(
            self.default_factory,
            zip(self.game_ids, self.raw_scores),
        )

    @classmethod
    def train(cls, ratings: pl.DataFrame) -> Self:
        """Train the recommender from ratings data."""
        raise NotImplementedError

    @classmethod
    def train_from_csv(cls, ratings_file: Path | str) -> Self:
        """Train the recommender from a ratings file in CSV format."""
        ratings = pl.read_csv(ratings_file)
        return cls.train(
            ratings.select(cls.id_field, cls.user_id_field, cls.rating_id_field),
        )

    @classmethod
    def train_from_json_lines(cls, ratings_file: Path | str) -> Self:
        """Train the recommender from a ratings file in JSON lines format."""
        ratings = pl.read_ndjson(ratings_file)
        return cls.train(
            ratings.select(cls.id_field, cls.user_id_field, cls.rating_id_field),
        )

    @property
    def known_games(self) -> frozenset[int]:
        if self._known_games is not None:
            return self._known_games
        self._known_games = frozenset(self.game_ids)
        return self._known_games

    @property
    def rated_games(self) -> frozenset[int]:
        return self.known_games

    @property
    def num_games(self) -> int:
        return len(self.game_ids)

    @property
    def known_users(self) -> frozenset[str]:
        return frozenset()

    def default_factory(self) -> float:
        """Default value for unknown games."""
        return self.default_value

    def _recommendation_scores(
        self,
        users: int,
        games: list[int] | None = None,
    ) -> np.ndarray:
        """Popularity scores."""
        scores = (
            np.array([self.scores[game_id] for game_id in games])
            if games
            else self.raw_scores
        )
        return np.tile(scores, [users, 1])

    def recommend(
        self,
        users: Iterable[str],
        **kwargs,  # noqa: ARG002
    ) -> pl.DataFrame:
        """Popular recommendations for certain users."""
        users = list(users)
        scores = self._recommendation_scores(users=len(users))
        return dataframe_from_scores(users=users, games=self.game_ids, scores=scores)

    def recommend_as_numpy(
        self,
        users: Iterable[str],
        games: Iterable[int],
    ) -> np.ndarray:
        """Popular recommendations for certain users and games as a numpy array."""
        users = list(users)
        games = list(games)
        return self._recommendation_scores(users=len(users), games=games)

    def recommend_group(
        self,
        users: Iterable[str],  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ) -> pl.DataFrame:
        """Popular recommendations for a group of users."""
        scores = self._recommendation_scores(users=1)
        return dataframe_from_scores(users=["_all"], games=self.game_ids, scores=scores)

    def recommend_group_as_numpy(
        self,
        users: Iterable[str],  # noqa: ARG002
        games: Iterable[int],
    ) -> np.ndarray:
        """Popular recommendations for a group of users and games as a numpy array."""
        games = list(games)
        return self._recommendation_scores(users=1, games=games)

    def recommend_similar(
        self,
        games: Iterable[int],
        **kwargs,
    ) -> pl.DataFrame:
        raise NotImplementedError

    def similar_games(
        self,
        games: Iterable[int],
        **kwargs,
    ) -> pl.DataFrame:
        raise NotImplementedError


class PopularMeanGamesRecommender(PopularGamesRecommender):
    """Recommend games by their mean rating score."""

    @classmethod
    def train(cls, ratings: pl.DataFrame) -> Self:
        data = ratings.group_by(cls.id_field).agg(
            mean=pl.col(cls.rating_id_field).mean(),
        )
        return cls(
            game_ids=data[cls.id_field],
            scores=data["mean"].to_numpy(),
            default_value=cast(float, ratings[cls.rating_id_field].mean()),
        )


class PopularBayesianGamesRecommender(PopularGamesRecommender):
    """Recommend games by their Bayesian average rating score."""

    ratings_per_dummy: float = 10_000
    dummy_rating: float | None = 5.5

    @classmethod
    def train(cls, ratings: pl.DataFrame) -> Self:
        num_dummies = len(ratings) / cls.ratings_per_dummy
        dummy_rating = cast(
            float,
            ratings[cls.rating_id_field].mean()
            if cls.dummy_rating is None
            else cls.dummy_rating,
        )

        stats = (
            ratings.group_by(cls.id_field)
            .agg(
                mean=pl.col(cls.rating_id_field).mean(),
                size=pl.len(),
            )
            .with_columns(
                bayes=pl.col("mean") * pl.col("size")
                + dummy_rating * num_dummies / (pl.col("size") + num_dummies),
            )
        )

        return cls(
            game_ids=stats[cls.id_field],
            scores=stats["bayes"].to_numpy(),
            default_value=dummy_rating,
        )


class PopularNumRatingsGamesRecommender(PopularGamesRecommender):
    """Recommend games by their number of ratings."""

    @classmethod
    def train(cls, ratings: pl.DataFrame) -> Self:
        data = ratings.group_by(cls.id_field).agg(count=pl.len())
        return cls(
            game_ids=data[cls.id_field],
            scores=data["count"].to_numpy(),
        )
